#include <joyos.h>
#include <math.h>

#define LEFTMOTORPORT (2) //red always faces down
#define RIGHTMOTORPORT (0) //red always faces up
#define MAX_VEL (255)//motor limit
#define MIN_COMPOSITE_VEL (150)
#define MIN_VEL_FORWARD (20)
#define DISTANCESPEEDRATIO (3)
#define THETA_UNITS_TO_DEGREES (0.08793)
#define GAIN (15)  //Last stable value: 10
#define VPS_TO_FEET (443.4)

//Begin includes from PathDrive

    //Values used in function

    //Both values currently in inches
    #define ROBOT_HEIGHT 10
    #define CAMERA_HEIGHT 113

    //New Structures

    typedef struct RobotPosition
    {
        int x;
        int y;
        int theta;
    }RobotPosition;

    typedef struct GoalPosition
    {
        int x;
        int y;
    }GoalPosition;

    //Functions for path creation

    RobotPosition getCurrentPosition();
    int getSimpleHeading(int, int, int, int);
    float getSimpleDistanceTo(int, int, int, int);
    void recalibrateGyro();
    int angleModulus(int);
    int rightAngleModulus(int);

//End includes


float limit_vel(int forward, int turn)
{
    //Are we turning fast enough?
	if(abs(forward) + abs(turn) > MIN_COMPOSITE_VEL)
        //Yes
    {
        if ((abs(forward) < MIN_VEL_FORWARD) && (abs(forward) > 2))
            //Contains a dead zone for staying still
        {
            forward = MIN_VEL_FORWARD * (forward / abs(forward));
        }

        int unlimitedVel = forward + turn;
        if (unlimitedVel > 255)             {return 255;}
        else if (unlimitedVel < -255)       {return -255;}
        else                                {return unlimitedVel;}
    }
    else
        //No, not turning fast enough
    {
        //Scale velocity up
        return ((forward + turn) * MIN_COMPOSITE_VEL / (abs(forward) + abs(turn)));
    }
}


void simpleDriveForward(int velocity)
{
	motor_set_vel(LEFTMOTORPORT, velocity);
	motor_set_vel(RIGHTMOTORPORT, velocity);

}

void goTowards(int forwardVel, int desiredHeading, int forwardsOrBackwards)
//must be in a while loop specifying how long motor's turned on
{
    int localForwardVel = forwardVel;
	int direction = angleModulus((int) gyro_get_degrees()); // -180 < degrees < 180
	int error = angleModulus(desiredHeading - (direction));

	//Makes robot go backwards if this is advantageous
	if(forwardsOrBackwards==0 && ((error>90) || (error<-90)))
	{
	    localForwardVel = -localForwardVel;
	    error = rightAngleModulus(error);
	}
    else if(forwardsOrBackwards==-1)
    {
        error = angleModulus(error + 180);
        localForwardVel = -localForwardVel;
    }

    /*if (error>150 || error<-150) //to prevent spinning around a point
    {
        if (localForwardVel>0)
            localForwardVel-=80;
        else
            localForwardVel+=80;
    }*/

	float leftVel = limit_vel(localForwardVel, -error*GAIN);
	float rightVel = limit_vel(localForwardVel, error*GAIN);
    motor_set_vel(LEFTMOTORPORT, leftVel);
	motor_set_vel(RIGHTMOTORPORT, rightVel);
}

void stop()
{
    motor_set_vel(LEFTMOTORPORT, 0);
    motor_set_vel(RIGHTMOTORPORT, 0);
}

void recalibrateGyro() //compare gyro_get_degrees() to VPS
{
    printf("Recalibrating \n");
    stop();
    pause(500);
	copy_objects();
	int theta = game.coords[0].theta;
    gyro_set_degrees((theta * THETA_UNITS_TO_DEGREES));
}


void proportionalDrive(int x, int y, int direction)
//Direction:
//0 - Any, 1 - Forwards only, -1 - Backwards only
{
    //calculating heading
	RobotPosition robot = getCurrentPosition();
	int robotheading = getSimpleHeading(robot.x, robot.y, x, y);
    float distanceto = getSimpleDistanceTo(robot.x, robot.y, x, y);

	//calculating velocity
    int robotvelocity;
	if(distanceto > MAX_VEL * DISTANCESPEEDRATIO)
	{
		robotvelocity = MAX_VEL;
	}
	else
	{
		robotvelocity = (int)(distanceto / DISTANCESPEEDRATIO);
	}

    //printf("Velocity: %d Heading: %d \n", robotvelocity, robotheading);
	//driving in the direction of said velocity and heading
	goTowards(robotvelocity, robotheading, direction);
}

void readPosition()
{
    RobotPosition me = getCurrentPosition();
    printf("%d, %d \n", me.x, me.y);
}

void shake()
{
    for(int i=0; i<5; i++)
    {
        motor_set_vel(LEFTMOTORPORT, -255);
        motor_set_vel(RIGHTMOTORPORT, -255);
        pause(100);
        motor_set_vel(LEFTMOTORPORT, 255);
        motor_set_vel(RIGHTMOTORPORT, 255);
        pause(100);
    }
}

void driveStraight(int x, int y, int tolerance)
//Drives forward in the initial direction, until it reaches
//the goal point.
{
    RobotPosition me = getCurrentPosition();
    int angle = ((int)gyro_get_degrees()) % 360;
    int distToGo = 10000;
    int32_t startTime = get_time();
    do
    {
        goTowards(60, angle, 1);
        me = getCurrentPosition();
        distToGo = getSimpleDistanceTo(me.x, me.y, x, y);
    }while((distToGo > tolerance) && ((get_time() - startTime)<3000));
}

bool goToPointInternal(int x, int y, int direction, int tolerance)
//DO NOT USE THIS FOR NAVIGATION - Use only goToPoint.
{
    int32_t startTime = get_time();
    int32_t checkTime = get_time();
    RobotPosition me = getCurrentPosition();
    RobotPosition lastPosition = me;
    int lastAngle = gyro_get_degrees();
    while (getSimpleDistanceTo(me.x, me.y, x, y) > tolerance)
    {
        proportionalDrive(x, y, direction);
        me = getCurrentPosition();
        //Over 4 seconds - timeout
        if(get_time() - startTime > 4000)
        {
            return false;
        }

        //Less than half an inch in half a second - timeout
        if(get_time() - checkTime > 500)
        {
            if(getSimpleDistanceTo(me.x, me.y, lastPosition.x, lastPosition.y) < VPS_TO_FEET / 24
               && abs(angleModulus(gyro_get_degrees() - lastAngle)) < 15)
            {
                    return false;
            }
            checkTime = get_time();
            lastPosition = getCurrentPosition();
            lastAngle = gyro_get_degrees();
        }
    }
    return true;
}

void backOut(int x, int y)
{
    RobotPosition me = getCurrentPosition();
    recalibrateGyro();
    int32_t backwardsTime = get_time();
    while (get_time() - backwardsTime < 500)
    {
        proportionalDrive(2*me.x - x, 2*me.y - y, 0);
    }
}



/*bool goToPoint(int x, int y, int direction, int tolerance, int timeout)

 This is the basic function for all navigation.  It takes in the
 destination coordinates, a direction, and a tolerance in VPS units.
 It includes a timeout, in case we get stuck or run into something.
 Direction:
 0 = either
 1 = forward only
 -1 = backward only
 2 = special condition: timeout already triggered
 3 = timeout triggered TWICE

{
    RobotPosition me = getCurrentPosition();
    int32_t startTime = get_time();
    RobotPosition lastPosition = me;
    int lastAngle = gyro_get_degrees();
    while (getSimpleDistanceTo(me.x, me.y, x, y) > tolerance)
    {
        if(direction == 2 || direction == 3)      {proportionalDrive(x, y, 0);}
        else                    {proportionalDrive(x, y, direction);}
        me = getCurrentPosition();

        //Check every second whether we've moved.
        if(get_time() - startTime <= 500)
            continue;

        if(getSimpleDistanceTo(me.x, me.y, lastPosition.x, lastPosition.y) < VPS_TO_FEET / 24
           && abs(angleModulus(gyro_get_degrees() - lastAngle)) < 15)
        {
            /*Nope, no movement.  We're stuck.
             Don't panic.  Go backwards a bit, then go to the midpoint.
             If that doesn't work, go to further midpoint.


            recalibrateGyro();
            int32_t backwardsTime = get_time();
            while (get_time() - backwardsTime < 500)
            {
                proportionalDrive(2*me.x - x, 2*me.y - y, 0);
            }
            me = getCurrentPosition();
            int terr = currentTerritory(me.x, me.y);
            if (direction==2)
            {
                //If this is the second timeout, go to the next midpt
                x = midpoints[(terr+1)%6][0];
                y = midpoints[(terr+1)%6][1];
                direction = 3;
            } else if (direction==3){
                //If this is the third timeout, go to the previous midpt
                int destinationTerr = terr - 1;
                if(destinationTerr < 0)     {destinationTerr += 6;}
                x = midpoints[destinationTerr][0];
                y = midpoints[destinationTerr][1];
                direction = 2;
            } else {
                //This is the first timeout.  Go to the current midpt
                goToPoint(midpoints[terr][0], midpoints[terr][1], 2, tolerance);
            }
        } //End if: collision detected
        lastPosition = getCurrentPosition();
        lastAngle = gyro_get_degrees();
        startTime = get_time();

    }//End navigation loop
    //Timeouts
    /*if (get_time() - startTime > 5000)
     {
     recalibrateGyro();
     int terr = currentTerritory(me.x, me.y);
     //Have we timed out already?
     if (direction==2)
     {
     //Yes, exactly once - Go to the next midpoint.
     goToPoint(midpoints[(terr+1)%6][0], midpoints[(terr+1)%6][1], 3, tolerance);

     } else if (direction==3)
     {//Yes, twice already.  Go to the previous midpoint.
     int destinationTerr = terr - 1;
     if(destinationTerr < 0)     {destinationTerr += 6;}
     goToPoint(midpoints[destinationTerr][0], midpoints[destinationTerr][1], 2, tolerance);
     } else {
     //No, go to the current midpoint.
     goToPoint(midpoints[terr][0], midpoints[terr][1], 2, tolerance);
     }
     //Either case, reset the timer after that detour is over
     startTime = get_time();
     }

}*/
